// RenderEngine class implementation
#include "pch.h"

#include "RenderEngine.h"
#include "util.h"

namespace nf::cli::render {
    RenderEngine::RenderEngine(std::shared_ptr<Window> window, DisplayConfig display)
        : m_window(window)
        , m_display(display)
        , m_instance()
        , m_surface()
        , m_physicalDevice()
        , m_queueFIGraphics()
        , m_queueFIPresent()
    {
        NFLog("Initializing render engine");
        m_window->setDisplay(m_display);
        m_window->show();

        createInstance();
        createSurface(m_window->getHandle());
        pickPhysicalDevice();
    }

    void RenderEngine::createInstance() {
        VkApplicationInfo appInfo = { VK_STRUCTURE_TYPE_APPLICATION_INFO };
        appInfo.apiVersion = VK_API_VERSION_1_0;

        VkInstanceCreateInfo instanceCI = { VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO };
        instanceCI.pApplicationInfo = &appInfo;
#ifdef _DEBUG
        const char* validationLayerName = "VK_LAYER_KHRONOS_validation";
        instanceCI.ppEnabledLayerNames = &validationLayerName;
        instanceCI.enabledLayerCount = 1;
#endif
        std::vector<const char*> instanceExtNames = { VK_KHR_SURFACE_EXTENSION_NAME, VK_KHR_WIN32_SURFACE_EXTENSION_NAME };
        instanceCI.ppEnabledExtensionNames = instanceExtNames.data();
        instanceCI.enabledExtensionCount = instanceExtNames.size();
        if (vkCreateInstance(&instanceCI, nullptr, &m_instance) != VK_SUCCESS)
            NFError("Could not create Vulkan instance");
    }

    void RenderEngine::createSurface(HWND window) {
        VkWin32SurfaceCreateInfoKHR surfaceCI = { VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR };
        surfaceCI.hwnd = window;
        if (vkCreateWin32SurfaceKHR(m_instance, &surfaceCI, nullptr, &m_surface) != VK_SUCCESS)
            NFError("Could not create Vulkan surface");
    }

    void RenderEngine::pickPhysicalDevice() {
        // TODO: Save GPU choice in DisplayConfig
        // First, get list of all GPUs
        uint32_t numPhysicalDevices = 0;
        vkEnumeratePhysicalDevices(m_instance, &numPhysicalDevices, nullptr);
        if (numPhysicalDevices == 0)
            NFError("No Vulkan GPUs found");

        std::vector<VkPhysicalDevice> physicalDevices(numPhysicalDevices);
        vkEnumeratePhysicalDevices(m_instance, &numPhysicalDevices, physicalDevices.data());

        // Then gather information on them and save first dedicated
        std::optional<int> firstDedicatedPhysicalDeviceIndex;
        std::vector<VkPhysicalDeviceProperties> physicalDeviceProperties;
        for (int i = 0; i < numPhysicalDevices; i++) {
            VkPhysicalDeviceProperties currentPhysicalDeviceProperties = {};
            vkGetPhysicalDeviceProperties(physicalDevices[i], &currentPhysicalDeviceProperties);
            if (currentPhysicalDeviceProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU && !firstDedicatedPhysicalDeviceIndex.has_value())
                firstDedicatedPhysicalDeviceIndex = i;

            physicalDeviceProperties.push_back(currentPhysicalDeviceProperties);
        }

        // Try dedicated first, then try every GPU in order
        if (firstDedicatedPhysicalDeviceIndex.has_value()) {
            int dedicatedIndex = firstDedicatedPhysicalDeviceIndex.value();
            physicalDevices.insert(physicalDevices.begin(), physicalDevices[dedicatedIndex]);
            physicalDevices.erase(physicalDevices.begin() + dedicatedIndex + 1);
            physicalDeviceProperties.insert(physicalDeviceProperties.begin(), physicalDeviceProperties[dedicatedIndex]);
            physicalDeviceProperties.erase(physicalDeviceProperties.begin() + dedicatedIndex + 1);
        }

        struct PhysicalDeviceQueueFamilyIndices {
            uint32_t graphics, present;
        };

        auto getQueueFamilyIndices = [&](int currentPhysicalDeviceIndex) -> std::optional<PhysicalDeviceQueueFamilyIndices> {
            uint32_t numQueueFamilies = 0;
            vkGetPhysicalDeviceQueueFamilyProperties(physicalDevices[currentPhysicalDeviceIndex], &numQueueFamilies, nullptr);
            std::vector<VkQueueFamilyProperties> queueFamilyProperties(numQueueFamilies);
            vkGetPhysicalDeviceQueueFamilyProperties(physicalDevices[currentPhysicalDeviceIndex], &numQueueFamilies, queueFamilyProperties.data());
            std::optional<uint32_t> graphicsIndex, presentIndex;

            for (int i = 0; i < numQueueFamilies; i++) {
                if (queueFamilyProperties[i].queueFlags & VK_QUEUE_GRAPHICS_BIT && !graphicsIndex.has_value())
                    graphicsIndex = i;

                VkBool32 hasPresentSupport = VK_FALSE;
                vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevices[currentPhysicalDeviceIndex], i, m_surface, &hasPresentSupport);
                if (hasPresentSupport && !presentIndex.has_value())
                    presentIndex = i;
            }

            if (graphicsIndex.has_value() && presentIndex.has_value())
                return PhysicalDeviceQueueFamilyIndices{ graphicsIndex.value(), presentIndex.value() };

            return std::nullopt;
        };

        std::optional<PhysicalDeviceQueueFamilyIndices> chosenPhysicalDeviceIndices;
        for (int i = 0; i < numPhysicalDevices; i++) {
            chosenPhysicalDeviceIndices = getQueueFamilyIndices(i);
            if (chosenPhysicalDeviceIndices.has_value()) {
                // GPU found!
                m_physicalDevice = physicalDevices[i];
                m_queueFIGraphics = chosenPhysicalDeviceIndices->graphics;
                m_queueFIPresent = chosenPhysicalDeviceIndices->present;

                NFLog(std::format("GPU - {}", physicalDeviceProperties[i].deviceName));
                break;
            }
        }

        if (m_physicalDevice == nullptr)
            NFError("No Vulkan GPUs were found to be compatible");
    }

    RenderEngine::~RenderEngine() {
        vkDestroySurfaceKHR(m_instance, m_surface, nullptr);
        vkDestroyInstance(m_instance, nullptr);
    }
}
